/* plain.c  1992 January 6 [gh]
+-----------------------------------------------------------------------------
| Abstract:
|    Plain filter module
|
| Authorship:
|    Copyright (c) 1987-1995 Gisle Hannemyr.
|    This program is free software;  you can redistribute it and/or modify
|    it under the terms of the GNU General Public License, as published by
|    the Free Software Foundation. See the file "copying.txt" for details.
|
|    If you intend to distribute changed versions of this module, please add
|    a comment below describing what you have changed/added and why, and email
|    the file to me.  I maintain the module, and shall appreciate copies of
|    bug fixes and improved versions.
|    My email address is: <gisle@hannemyr.no>.
|
| Access programs:
|    inittable : initialize default transformation table
|    listtable : list paths to search for tables
|    readtable : initialize transformation table from file
|    doplain   : interprete one file
|
| See main module for history and comments.
+---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "pep.h"

/*---( defines )------------------------------------------------------------*/

#ifdef __MSDOS__
#define LIBY	"D:\\lib\\pep"
#else
#define LIBY	"/local/share/lib/pep"
#endif
#define WSHH        0x1e                /* WordStar invisible soft hyphen.  */
#define WSSH        0x1f                /* WordStar visible   soft hyphen.  */
#define WSSC        0x8d                /* WordStar soft carrige return.    */
#define WSSS        0xa0                /* WordStar soft space.             */

#ifndef	BUFSIZE
#define BUFSIZE	8192
#endif	/* BUFSIZE */


/*---( types )--------------------------------------------------------------*/

typedef int FOLDMATRIX[11][2];


/*---( constants )----------------------------------------------------------*/

/* Test for Swedish characters too, but after the Norwegian ones. */
static FOLDMATRIX decfold = {
   { 198,  91 },    /* AE [ (N) */
   { 216,  92 },    /* OE \ (N) */
   { 197,  93 },    /* AA ] (-) */
   { 230, 123 },    /* ae { (N) */
   { 248, 124 },    /* oe | (N) */
   { 229, 125 },    /* aa } (-) */
   { 196,  91 },    /* AE [ (S) */
   { 214,  92 },    /* OE \ (S) */
   { 228, 123 },    /* ae { (S) */
   { 246, 124 },    /* oe | (S) */
   {   0,   0 }
};


static FOLDMATRIX ibmfold = {
   { 146,  91 },    /* AE [ (N) */
   { 157,  92 },    /* OE \ (N) */
   { 143,  93 },    /* AA ] (-) */
   { 145, 123 },    /* ae { (N) */
   { 155, 124 },    /* oe | (N) */
   { 134, 125 },    /* aa } (-) */
   { 142,  91 },    /* AE [ (S) */
   { 153,  92 },    /* OE \ (S) */
   { 132, 123 },    /* ae { (S) */
   { 148, 124 },    /* oe | (S) */
   {   0,   0 }
};


static FOLDMATRIX macfold = {
   { 174,  91 },    /* AE [ (N) */
   { 175,  92 },    /* OE \ (N) */
   { 129,  93 },    /* AA ] (-) */
   { 190, 123 },    /* ae { (N) */
   { 191, 124 },    /* oe | (N) */
   { 140, 125 },    /* aa } (-) */
   { 128,  91 },    /* AE [ (S) */
   { 133,  92 },    /* OE \ (S) */
   { 138, 123 },    /* ae { (S) */
   { 154, 124 },    /* oe | (S) */
   {   0,   0 }
};


/*---( variables )----------------------------------------------------------*/

static unsigned char Buffer[BUFSIZE];       /* Output buffer                */
static unsigned char CTable[256];           /* General fold matrix          */

static int  BuffSs = 0;                     /* Bona-fide chars. in string.  */
static int  BuffXx = 0;                     /* Horisontal pos. in buffer.   */
static int  NSpace = 0;			    /* No. of spaces not flushed.   */


/*---( forward )------------------------------------------------------------*/

char *getenv();


/*---( housekeeping )-------------------------------------------------------*/

/*
| Abs: Is it a control character?  (According to pep's rather complicated
|      concept of such.)
| Ret: TRUE if it is, else FALSE (pep hacks ctrl. chars, leave others alone).
*/
static BOOL ctrlp(int cc)
{
   if (isascii(cc)) {
      if (wflag1 && (cc == WSSH)) return(FALSE);
      if (iscntrl(cc)) return(TRUE); else return(FALSE);
   } else if (!bflagb) return(FALSE);
   /* Assert: if it was 7 bit ASCII, or if the upper 128 set shall be       */
   /* considered legal character, then we have returned by now.             */

   if (wflag1 && (cc == WSSC)) return(FALSE);
   /* Assert: if it was WordStar soft CR, then we have returned by now.     */

   if (iflagi && IFrst) {             /* We are folding to IBM charset,     */
      int jj = 0;
      while (ibmfold[jj][0] && (ibmfold[jj][0] != cc)) jj++;
      return(!ibmfold[jj][0]);        /*    so those are not control chars. */
   }
   return(TRUE);
   /* If all else fails, it must be a control character.                    */
} /* ctrlp */


/*---( transformation table )-----------------------------------------------*/

/*
| Abs: Catenate tabledir and cname to create a full searchpath.
| Ret: Pointer to the full, catenated path.
*/
static char *findpath(char *tabledir, char *cname)
{
   static char fullpath[1024];
   char *ss;

   if (!tabledir) return(cname); /* Bail out under VMS, etc.   */
   ss = fullpath;
   while (*ss++ = *tabledir++) ; /* strcpy(fullpath,tabledir); */
   ss--;                         /* ss points to terminator    */
   if (ss[-1] != DIRCHAR) { *ss = DIRCHAR; *++ss = '\0'; }
   while (*ss++ = *cname++) ;    /* strcat(fullpath,cname); */
   return(fullpath);
} /* findpath */


/*
| Abs: Initialize default transformation table.
*/
void inittable(void)
{
   int cc;

   for (cc = 0; cc < 256; cc++) CTable[cc] = cc;
} /* inittable */



/*
| Abs: List paths to search for tables.
| Par: tabledir = startup directory (or NULL)
*/
void listtable(char *tabledir)
{
    char *ss;
    int ii = 2;

    fputs("\n\
The following directories are searched for conversion tables:\n\
    1: . <current directory>\n", stderr);
    if (tabledir) fprintf(stderr, "%5d: %s\n", ii++, tabledir);
    if (ss = getenv("PEP")) fprintf(stderr, "%5d: %s\n", ii++, ss);
    fprintf(stderr, "%5d: %s\n", ii++, LIBY);
} /* listtable */


/*
| Abs: Initialize transformation table from file.
| Par: tabledir = pointer to table to initialize
|      cname    = file to read table from
|      echo     = TRUE to echo comments to stderr, else quiet.
*/
void readtable(char *tabledir, char *cname, BOOL echo)
{
   FILE *fdt;
   char *ss;

   if ((fdt = fopen(cname,"r")) == NULL) {
      if (tabledir)
         fdt = fopen(findpath(tabledir,cname),"r");
      if (fdt == NULL) {
         if (tabledir = getenv("PEP"))
            fdt = fopen(findpath(tabledir,cname),"r");
         if (fdt == NULL) {
            fdt = fopen(findpath(LIBY,cname),"r");
            if (fdt == NULL) {
               fprintf(stderr,"can't open translation table \"%s\"\n", cname);
               exit(ERROR_EXIT);
            } /* if file not found in hardwired directory */
         } /* if file not found in environment directory */
      } /* if file not found in startup directory */
   } /* if file not found in local directory */

   while (fgets(Buffer, BUFSIZE-1, fdt)) {
      char *ss;
      int ii;
      if (ss = strchr(Buffer,'#')) {
	  if (echo && (ss[1] != '#')) {
             fputs("   ",stderr);
             fputs(ss,stderr);
          }
          *ss = '\0';
      } /* if comment */
      for (ii = strlen(Buffer) - 1; isspace(Buffer[ii]) && (ii >= 0); ii--);
      Buffer[++ii] = 0;
      if (*Buffer) {
         int tt, ff;
         if (sscanf(Buffer,"%d %d",&ff,&tt) != 2) mess(6);
         if ((tt < 0) || (tt > 255) || (ff < 0) || (ff > 255)) mess(6);
         /* fprintf(stderr,"[%s] %d <== %d\n",Buffer,tt,ff); */
         CTable[ff] = tt;
      }
   } /* while */
   fclose(fdt);
} /* readtable */


/*---( foldings )-----------------------------------------------------------*/

/*
| Abs:
*/
static unsigned char fold8(int cc, FOLDMATRIX fold)
{
   if (cc >= ILimit) {
      int jj = 0;
      while (fold[jj][IFrst] && (fold[jj][IFrst] != cc)) jj++;
      if (fold[jj][IFrst]) cc = fold[jj][ILast];
   }
   return(cc);
} /* fold8 */


/*
| Abs: Is character white space.
| Des: Turbo-C return bogus values if isspace is called with arg > 128.
| Ret: TRUE if space, else FALSE.
*/
static BOOL isspace8(unsigned char cc)
{
    return(!(cc & 0x80) && isspace(cc));
} /* isspace8 */


/*
| Abs: Flush accumulated whitespace, compressing spaces into tabs.
*/
static void flushspace(void)
{
   int ii;

   if (cflagc && OTabSz) {
      while (NSpace > 1 && NSpace >= (ii = OTabSz - (LineXx % OTabSz))) {
      	 Buffer[BuffXx++] = '\t';
         NSpace -= ii;
         LineXx += ii;
      } /* while */
   } /* if compressing tabs */
   LineXx += NSpace;
   while (NSpace--) Buffer[BuffXx++] = ' ';
   NSpace = 0;
} /* flushspace */


/*
| Abs: Flush and terminate a single output line.
| Des: Removes trailing spaces, except if the -v option is active. In that case
|      each paragraph gets one added (this is a slight cosmetic bug.)
*/
static void flusheoln(unsigned char *ss)
{
    int xx;

    if (ss) {
	xx = strlen(ss);
	while (isspace8(ss[xx-1]) && xx) xx--;
	ss[xx] = '\0';
	/* printf("[%s]\n", ss); */
	fputs(ss, Fdo);
    } /* if */

    if (wflag1) {
	if (BuffXx) putc(WSSC,Fdo); else putc('\r',Fdo);
	putc('\n',Fdo);
    } else {
	if ((vflagv && BuffXx) || (EndOLn == -2)) putc(' ',Fdo);
	else {
	    if (EndOLn == -1) { putc('\r',Fdo); putc('\n',Fdo); }
	    else putc(EndOLn,Fdo);
	} /* if */
    } /* if */
} /* flusheoln */


/*
| Abs: Flush line in buffer to the output file.
| Par: igps = set if we should ignore trailing spaces.
| Sef: Zero NSpace, BuffXx and LineXx counts.
*/
static void flushline(BOOL igsp)
{
    unsigned char *ss;
    unsigned char *tt;
    int ll, nosp;

    if (!sflags || (BuffSs >= StrSiz)) {
	/* printf("[%d: %s]\n", igsp, Buffer); */
	/* Remove trailing spaces */
	nosp = 0;
	while (isspace8(Buffer[BuffXx-1]) && BuffXx) { BuffXx--; nosp++; }
	Buffer[BuffXx] = '\0';

	if (wflag1) {
	    int xx;
	    if (BuffXx && (Buffer[BuffXx-1] == '-')) Buffer[BuffXx-1] = WSSH;
	    xx = 0;
	    while (Buffer[xx]) {
		if (Buffer[xx] == ' ') Buffer[xx] = WSSS;
		xx++;
	    } /* while */
	} /* if to WordStar */

	if (LinSiz) {
	    ss = Buffer;
	    while (strlen(ss) > LinSiz) {
		tt = ss + LinSiz;
		if (isspace8(*tt)) {
		    while (*tt && isspace8(*tt)) tt++;
		    tt--;
		} else {
		    while (!isspace8(*tt) && tt != ss) tt--;
		} /* if */
		if (tt == ss) {
		    while (*tt && !isspace8(*tt)) tt++;
		} /* no room for word */
		if (!(*tt)) break;
		*tt = '\0';
		flusheoln(ss);
		ss = tt + 1;
	    } /* while */
	    flusheoln(ss);
	    if (lflagl) flusheoln(NULL);
	} else flusheoln(Buffer);
	if (!igsp) {
	    for (ll = 0; ll < nosp; ll++) putc(' ',Fdo);
	} /* if */
	showprogress();
    } /* if enough of a string to print it */
    NSpace = BuffXx = LineXx = 0;
} /* flushline */


/*
| Abs: Put character into line buffer.
| Des: First, make some simple character foldings.  Then, put the character
|      into a line buffer.  A control character may be converted to hex,
|      surrounded by angle brackets.
*/
static void putline(int cc)
{
   static BOOL wasbs = FALSE;
   static BOOL wascr = FALSE;
   static BOOL wasoe = FALSE;
   BOOL        isaoe = FALSE;

   if (BuffXx >= (BUFSIZE-5)) flushline(TRUE);      /* Panic!                    */
   if (zflagz) cc = cc & 0x7f;                      /* Fold to 7 bit.            */
   if (mflagm && !IFrst && (cc == '\r')) cc = '\n'; /* Mac uses CR as terminator */
   /* This is a hack to fool the stuff below who removes redundant CR's.         */
   /* When we are converting from Mac format, we want to keep them all.          */

   if (wasoe) {
      wasoe = FALSE;
      if (cc !=  92) {
	 Buffer[BuffXx] = '\\';
	 LineXx++;
	 BuffXx++;
      } /* if */
   } else if ((cc ==  92) && kflagk && IFrst) { wasoe++; return; }

   if (dflagd) { if (!IFrst && (cc == 216)) isaoe++; cc = fold8(cc,decfold); }
   if (iflagi) { if (!IFrst && (cc == 157)) isaoe++; cc = fold8(cc,ibmfold); }
   if (mflagm) { if (!IFrst && (cc == 175)) isaoe++; cc = fold8(cc,macfold); }
   if (gflagg) cc = CTable[cc];               /* Fold from table.           */
   if (wflag0) {
      if (cc == WSHH) return;                 /* WS invisible soft hyphen.  */
      if (cc == WSSH) cc = '-';               /* WS   visible soft hyphen.  */
   }

   if (wascr && (cc != ' ') && (cc != '\t')&& (cc != '\r')) {
      wascr = FALSE;
      /*printf("[BuffXx: %d]\n", BuffXx); */
      if (BuffXx) {
         flushline((BOOL)cc == '\n');
         if (cc == '\n') return;
      } /* if anything in buffer */
   } /* if wascr */

   if     ((cc == '\b') && BuffXx) { BuffXx--; wasbs++; }
   else if (cc == '\f') {
      if (BuffXx) flushline(TRUE);
      if (wflag1) fputs(".PA\n",Fdo);
      else if (!(bflagb + sflags)) putc('\f',Fdo);
      wasbs = FALSE;
   }
   else if (cc == '\n') { flushline(TRUE); BuffSs++; }
   else if (cc == '\r') { wascr++; }
   else if (cc == '\t') {
      if (tflagt || cflagc) NSpace += ITabSz - ((LineXx + NSpace) % ITabSz);
      else { Buffer[BuffXx++] = cc; LineXx++; }
      BuffSs++;
      wasbs = FALSE;
   }
   else if (ctrlp(cc)) {
      if (xflagx) {                       /* Expanding control chars. */
         char *ss;
         if (NSpace) flushspace();
         ss = (char *)&Buffer[BuffXx];
         sprintf(ss,"<%02xh>",cc);
         BuffXx += 5;
         LineXx += 5;
      }
      else if (BuffXx) flushline(TRUE); /* Removing  control chars. */
      BuffSs = 0;
      wasbs  = FALSE;
   }
   else if (cflagc && cc == ' ') NSpace++;
   else {
      if (NSpace) flushspace();
      if (!wasbs || (cc != '_')) Buffer[BuffXx] = cc;
      LineXx++;
      BuffXx++;
      if (kflagk && isaoe) {
	 Buffer[BuffXx] = cc;
	 LineXx++;
	 BuffXx++;
      }
      BuffSs++;
      wasbs = FALSE;
   }
} /* putline */


/*---( file loop )----------------------------------------------------------*/

/*
| Abs: Read (and write) one complete plain file.
*/
void doplain(void)
{
   int cc;

   while ((cc = getc(Fdi)) != EOF) putline(cc);
   if (BuffXx) flushline(TRUE);
} /* doplain */

/* EOF */
